#ifndef MSP_TEST_TEST_H_
#define MSP_TEST_TEST_H_

#include <list>
#include <string>
#include <msp/strings/format.h>
#include "function.h"

namespace Msp {
namespace Test {

#define EXPECT(c) expect(c, #c)
#define EXPECT_EQUAL(a, b) expect_equal(a, a==b, #a " == " #b)

class test_failed: public std::exception
{
private:
	std::string what_;

public:
	test_failed(const std::string &w): what_(w) { }
	virtual ~test_failed() throw() { }

	const char *what() const throw() { return what_.c_str(); }
};


class Test
{
public:
	class Factory
	{
	protected:
		Factory();

	public:
		virtual const char *get_name() const = 0;
		virtual Test *create() const = 0;
	};

private:
	std::string name;
	std::list<Function *> functions;
	std::string detail_info;
	std::string detail_debug;
	bool passed;

protected:
	Test(const std::string &);
public:
	virtual ~Test();

	static std::list<Factory *> &get_factories();

protected:
	template<typename T>
	Function &add(void (T::*f)(), const std::string &d)
	{
		functions.push_back(new TypedFunction<T>(f, d));
		return *functions.back();
	}

public:
	static void run_single(const std::string &, bool);
	static void run_all(bool);

	void run(bool);
private:
	void start_test(const std::string &);
	void pass_test();
	void fail_test(const std::string &);

protected:
	void expect(bool, const std::string &);

	template<typename T>
	void expect_equal(const T &var, bool cond, const std::string &expr)
	{
		if(!cond)
			throw test_failed(format("!(%s)\nActual: %s", expr, var));
		debug(expr);
	}

	void info(const std::string &);
	void debug(const std::string &);
};

template<typename T>
class RegisteredTest: public Test
{
private:
	class Factory: public Test::Factory
	{
	public:
		virtual const char *get_name() const { return T::get_name(); }
		virtual Test *create() const { return new T; }
	};

	static Factory factory;

protected:
	RegisteredTest(): Test(T::get_name()) { (void)factory; }
};

template<typename T>
typename RegisteredTest<T>::Factory RegisteredTest<T>::factory;

} // namespace Test
} // namespace Msp

#endif
